#include "caldav.h"

CURL *curl;
char access_token[sizeof(char)*2048] = "";
char refresh_token[sizeof(char)*512] = "";
int token_ttl = 0;

// Local bind address for receiving OAuth callbacks.
// Reserve 2 chars for the ipv6 square brackets.
char ip[INET6_ADDRSTRLEN], ipstr[INET6_ADDRSTRLEN + 2];

/* Write a random code challenge of size len to dest */
void random_code_challenge(size_t len, char* dest) {
    // https://developers.google.com/identity/protocols/oauth2/native-app#create-code-challenge
    // A code_verifier is a random string using characters [A-Z] / [a-z] / [0-9] / "-" / "." / "_" / "~"
    srand(time(NULL));

    char alphabet[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~";
    size_t alphabet_size = strlen(alphabet);

    int i = 0;
    for (i = 0; i < len; i++) {
        dest[i] = alphabet[rand() % alphabet_size];
    }
    dest[len] = '\0';
}

static size_t curl_write_mem_callback(void * contents, size_t size, size_t nmemb, void *userp) {
    size_t realsize = size * nmemb;
    struct curl_mem_chunk* mem = (struct curl_mem_chunk*)userp;

    char* ptr = realloc(mem->memory, mem->size + realsize + 1);
    if (!ptr) {
        tracepoint(diary, error, "Not enough memory (realloc in CURLOPT_WRITEFUNCTION returned NULL)");
        return 0;
    }

    mem->memory = ptr;
    memcpy(&(mem->memory[mem->size]), contents, realsize);
    mem->size += realsize;
    mem->memory[mem->size] = 0;

    return realsize;
}

// todo
// https://beej.us/guide/bgnet/html
void* get_in_addr(struct sockaddr *sa) {
    if (sa->sa_family == AF_INET) {
        return &(((struct sockaddr_in*)sa)->sin_addr);
    }

    return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

/*
* Extract OAuth2 code from http header. Reads the code from
* the http header in http_header and writes the OAuth2 code
* to the char string pointed to by code.
*/
char* extract_oauth_code(char* http_header) {
    // example token: "/?code=&scope="
    char* res = strtok(http_header, " ");
    while (res != NULL) {
        if (strstr(res, "code") != NULL) {
            res = strtok(res, "="); // code key
            res = strtok(NULL, "&"); // code value
            break;
        }
        res = strtok(NULL, " ");
    }
    return res;
}

/* Read access token from file and refresh global token vars */
char* read_tokenfile() {
    FILE* token_file;
    char* token_buf;
    long token_bytes;

    char* tokenfile_path = expand_path(CONFIG.google_tokenfile);
    token_file = fopen(tokenfile_path, "r");
    free(tokenfile_path);

    if (token_file == NULL) {
        perror("Warning - failed to open tokenfile in read_tokenfile()");
        return NULL;
    }

    fseek(token_file, 0, SEEK_END);
    token_bytes = ftell(token_file);
    rewind(token_file);

    token_buf = malloc(token_bytes + 1);
    if (token_buf != NULL) {
        size_t items_read = fread(token_buf, sizeof(char), token_bytes, token_file);
        if (items_read != token_bytes) {
            perror("Error while reading in read_tokenfile()");
            fclose(token_file);
            return NULL;
        }
        token_buf[token_bytes] = '\0';

        update_global_token_vars(token_buf);
    } else {
        perror("malloc failed");
        fclose(token_file);
        return NULL;
    }
    fclose(token_file);
    return token_buf;
}

/* Refresh global token variables from input buffer */
void update_global_token_vars(char* input) {
        char* new_access_token = extract_json_value(input, "access_token", true);
        tracepoint(diary, debug_string, "New access token", new_access_token);
        if (new_access_token != NULL) {
            strncpy(access_token, new_access_token, 250);
            free(new_access_token);
        }

        // program segfaults if NULL value is provided to atoi
        char* token_ttl_str = extract_json_value(input, "expires_in", false);
        if (token_ttl_str == NULL) {
            token_ttl = 0;
        } else {
            token_ttl = atoi(token_ttl_str);
        }
        free(token_ttl_str);

        // only update the existing refresh token if the request actually
        // contained a valid refresh_token, i.e, if it was the initial
        // interactive authZ request from token code confirmed by the user
        char* new_refresh_token = extract_json_value(input, "refresh_token", true);
        tracepoint(diary, debug_string, "New refresh token", new_refresh_token);
        if (new_refresh_token != NULL) {
            strncpy(refresh_token, new_refresh_token, 250);
            free(new_refresh_token);
        }
}

/* Persist contents from global token file variables to tokenfile */
void update_tokenfile_from_global_vars() {
    char* tokenfile_path = expand_path(CONFIG.google_tokenfile);
    FILE* tokenfile = fopen(tokenfile_path, "wb");

    if (tokenfile == NULL) {
        perror("Warning - failed to open tokenfile in update_tokenfile_from_global_vars()");
    } else {
        char contents[sizeof(char)*2048 + sizeof(char)*512 + 200];
        char* tokenfile_contents = "{\n"
        "  \"access_token\": \"%s\",\n"
        "  \"expires_in\": %i,\n"
        "  \"refresh_token\": \"%s\"\n"
        "}\n";
        sprintf(contents, tokenfile_contents,
                access_token,
                token_ttl,
                // Make sure the refresh token is re-written and persistet
                // to the tokenfile for further requests, because this token
                // is not returned by the refresh_token call:
                // https://developers.google.com/identity/protocols/oauth2/native-app#offline
                refresh_token);
        fputs(contents, tokenfile);
    }
    fclose(tokenfile);

    if (chmod(tokenfile_path, S_IRUSR|S_IWUSR) == -1) {
        perror("chmod");
    }

    free(tokenfile_path);
}

/* Get access token with authorization code or refresh token */
void get_access_token(char* code, char* verifier, bool refresh) {
    CURLcode res;
    char* postfields_refresh = "client_id=%s&client_secret=%s&grant_type=refresh_token&refresh_token=%s";
    char* postfields_code = "client_id=%s&client_secret=%s&code=%s&code_verifier=%s&grant_type=authorization_code&redirect_uri=http://%s:%i";
    char* postfields;

    tracepoint(diary, debug, "Getting access token get_access_token()");

    if (refresh) {
        // work with refresh token, subtract 3 placholders
        tracepoint(diary, debug_sizet, "strlen(refresh_token)", strlen(refresh_token));
        postfields = calloc(strlen(postfields_refresh) - 6 + strlen(CONFIG.google_clientid) + strlen(CONFIG.google_secretid) + strlen(refresh_token) + 1, sizeof(char));
        sprintf(postfields, postfields_refresh,
                CONFIG.google_clientid,
                CONFIG.google_secretid,
                refresh_token);
    } else {
        // work with auth code, subtract 6 placeholders
        tracepoint(diary, debug_sizet, "strlen(postfields_code)", strlen(postfields_code));
        tracepoint(diary, debug_sizet, "strlen(CONFIG.google_clientid)", strlen(CONFIG.google_clientid));
        tracepoint(diary, debug_sizet, "strlen(CONFIG.google_secretid))", strlen(CONFIG.google_secretid));
        tracepoint(diary, debug_sizet, "strlen(code)", strlen(code));
        tracepoint(diary, debug_sizet, "strlen(verifier)", strlen(verifier));
        tracepoint(diary, debug_sizet, "strlen(ipstr)", strlen(ipstr));
        tracepoint(diary, debug_sizet, "strlen(ipstr)", strlen(MKSTR(GOOGLE_OAUTH_REDIRECT_PORT)));
        postfields = calloc(strlen(postfields_code) - 12
                            + strlen(CONFIG.google_clientid)
                            + strlen(CONFIG.google_secretid)
                            + strlen(code)
                            + strlen(verifier)
                            + strlen(ipstr)
                            + strlen(MKSTR(GOOGLE_OAUTH_REDIRECT_PORT)) + 1, sizeof(char));
        sprintf(postfields, postfields_code,
                CONFIG.google_clientid,
                CONFIG.google_secretid,
                code,
                verifier,
                ipstr,
                GOOGLE_OAUTH_REDIRECT_PORT);
    }

    curl = curl_easy_init();

    // https://curl.se/libcurl/c/getinmemory.html
    struct curl_mem_chunk token_resp;
    token_resp.memory = malloc(1);
    if (token_resp.memory == NULL) {
        perror("malloc failed");
        return;
    }
    token_resp.size = 0;

    if (curl) {
        curl_easy_setopt(curl, CURLOPT_URL, GOOGLE_OAUTH_TOKEN_URL);
        curl_easy_setopt(curl, CURLOPT_POSTFIELDS, postfields);
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_write_mem_callback);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *)&token_resp);

        res = curl_easy_perform(curl);

        curl_easy_cleanup(curl);

        // Only write tokenfile if the token in token_resp.memory looks valid.
        // Assume it's valid, when the curl request succeeds (no HTTP errors).
        // Otherwise, a valid token in the possisbly already existing file would
        // be overwritten. This happens usually during syncing (s/S) while offline.
        if (res == CURLE_OK) {
            // update global token variables
            update_global_token_vars(token_resp.memory);
            update_tokenfile_from_global_vars();
        } else {
            tracepoint(diary, error_string, "curl_easy_perform() failed in get_access_token()", curl_easy_strerror(res));
        }
    }

    free(token_resp.memory);
    free(postfields);
}

/* Perform challenge/request to fetch authorization code */
char* get_oauth_code(const char* verifier, WINDOW* header) {
    struct addrinfo hints, *addr_res;

    memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    int status;
    if ((status=getaddrinfo(NULL, MKSTR(GOOGLE_OAUTH_REDIRECT_PORT), &hints, &addr_res)) != 0) {
        tracepoint(diary, error_string, "getaddrinfo() failed in get_oauth_code()", gai_strerror(status));
    }

    void *addr;
    char *ipver;
    //todo: extract
    //addr = get_in_addr(addr_res->ai_addr);
    if (addr_res->ai_family == AF_INET) {
        struct sockaddr_in *ipv4 = (struct sockaddr_in *) addr_res->ai_addr;
        addr = &(ipv4->sin_addr);
        ipver = "IPv4";
    } else { // IPv6
        struct sockaddr_in6 *ipv6 = (struct sockaddr_in6 *) addr_res->ai_addr;
        addr = &(ipv6->sin6_addr);
        ipver = "IPv6";
    }

    inet_ntop(addr_res->ai_family, addr, ip, sizeof ip);
    if (strcmp("IPv6", ipver) == 0) {
        sprintf(ipstr, "[%s]", ip);
    }

    // Show Google OAuth URI
    char uri[500];
    sprintf(uri, "%s?scope=%s&code_challenge=%s&response_type=%s&redirect_uri=http://%s:%i&client_id=%s",
            GOOGLE_OAUTH_AUTHZ_URL,
            GOOGLE_OAUTH_SCOPE,
            verifier,
            GOOGLE_OAUTH_RESPONSE_TYPE,
            ipstr,
            GOOGLE_OAUTH_REDIRECT_PORT,
            CONFIG.google_clientid);

    // Show the Google OAuth2 authorization URI in the header
    wclear(header);
    wresize(header, LINES, getmaxx(header));
    mvwprintw(header, 0, 0, "Go to Google OAuth2 authorization URI. Use 'q' or 'Ctrl+c' to quit authorization process.\n%s", uri);
    wrefresh(header);

    int socketfd = socket(addr_res->ai_family, addr_res->ai_socktype, addr_res->ai_protocol);
    if (socketfd < 0) {
       perror("Error opening socket");
    }

    // reuse socket address
    int yes=1;
    setsockopt(socketfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof yes);

    if (bind(socketfd, addr_res->ai_addr, addr_res->ai_addrlen) < 0) {
       perror("Error binding socket");
    }

    freeaddrinfo(addr_res);

    int ls = listen(socketfd, GOOGLE_OAUTH_REDIRECT_SOCKET_BACKLOG);
    if (ls < 0) {
       perror("Listen error");
    }

    struct pollfd pfds[2];

    pfds[0].fd = STDIN_FILENO;
    pfds[0].events = POLLIN;
    pfds[1].fd = socketfd;
    pfds[1].events = POLLIN;
    int fd_count = 2;
            
    int connfd, bytes_rec = 0, bytes_sent = 0;
    char http_header[8*1024];
    char* reply =
    "HTTP/1.1 200 OK\n"
    "Content-Type: text/html\n"
    "Connection: close\n\n"
    "<html>"
    "<head><title>Authorization successfull</title></head>"
    "<body>"
    "<p><b>Authorization successfull.</b></p>"
    "<p>You consented that diary can access your Google calendar.<br/>"
    "Please close this window and return to diary.</p>"
    "</body>"
    "</html>";

    // Handle descriptors read-to-read (POLLIN),
    // stdin or server socker, whichever is first
    for (;;) {
        int poll_count = poll(pfds, fd_count, -1);

        if (poll_count == -1) {
            perror("Erro in poll");
            break;
        }

        // Cancel through stdin
        if (pfds[0].revents & POLLIN) {
            noecho();
            int ch = wgetch(header);
            echo();
            // sudo showkey -a 
            // Ctrl+c: ^C 0x03
            // q     :  q 0x71
            if (ch == 0x03 || ch == 0x71) {
                tracepoint(diary, debug, "Hanging up, closing server socket");
                break;
            }
        }
        if (pfds[1].revents & POLLIN) {
            // accept connections but ignore client addr
            connfd = accept(socketfd, NULL, NULL);
            if (connfd < 0) {
               perror("Error accepting connection");
               break;
            }

            bytes_rec = recv(connfd, http_header, sizeof http_header, 0);
            if (bytes_rec < 0) {
                perror("Error reading stream message");
                break;
            }

            bytes_sent = send(connfd, reply, strlen(reply), 0);
            if (bytes_sent < 0) {
               perror("Error sending");
            }

            close(connfd);
            break;
        }
    } // end for ;;

    // close server socket
    close(pfds[1].fd);

    if (bytes_rec == 0) {
        tracepoint(diary, error, "No OAuth header found");
        return NULL;
    }

    char* code = extract_oauth_code(http_header);
    if (code == NULL) {
        tracepoint(diary, error, "No OAuth code in http header");
        return NULL;
    }

    char* oauth_code = malloc(strlen(code) + 1);
    strcpy(oauth_code, code);

    return oauth_code;
}

/* Make a CalDAV request */
char* caldav_req(struct tm* date, char* url, char* http_method, char* postfields, int depth, bool basicauth, char* content_type, long num_redirects) {
    // only support depths 0 or 1
    if (depth < 0 || depth > 1) {
        return NULL;
    }

    CURLcode res;

    curl = curl_easy_init();

    // https://curl.se/libcurl/c/getinmemory.html
    struct curl_mem_chunk caldav_resp;
    caldav_resp.memory = malloc(1);
    if (caldav_resp.memory == NULL) {
        perror("malloc failed");
        return NULL;
    }
    caldav_resp.size = 0;

    if (curl) {
        // fail if not authenticated, !CURLE_OK
        curl_easy_setopt(curl, CURLOPT_FAILONERROR, 1L);
        //curl_easy_setopt(curl, CURLOPT_VERBOSE, 1);
        curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, http_method);
        tracepoint(diary, debug_string, "curl_easy_perform() request url", url);
        curl_easy_setopt(curl, CURLOPT_URL, url);
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_write_mem_callback);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *)&caldav_resp);

        // construct header fields
        struct curl_slist *header = NULL;

        // default to basic auth when Google credentials are not set
        if (basicauth) {
                char basicauth [strlen(CONFIG.caldav_username) + strlen(CONFIG.caldav_password) + 2];
                sprintf(basicauth, "%s:%s", CONFIG.caldav_username, CONFIG.caldav_password);
                curl_easy_setopt(curl, CURLOPT_USERPWD, basicauth);
                tracepoint(diary, debug_string, "curl_easy_perform() basicauth", basicauth);
        } else {
                char bearer_token[strlen("Authorization: Bearer ") + strlen(access_token) + 1];
                sprintf(bearer_token, "Authorization: Bearer %s", access_token);
                header = curl_slist_append(header, bearer_token);
        }

        char depth_header[strlen("Depth: 0") + 1];
        sprintf(depth_header, "Depth: %i", depth);
        header = curl_slist_append(header, depth_header);
        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, header);

        if (strcmp(content_type, "") != 0 && content_type != NULL) {
                char contentType[strlen("Content-Type: ") + strlen(content_type) + 1];
                sprintf(contentType, "Content-Type: %s", content_type);
                header = curl_slist_append(header, contentType);
        }

        // set postfields, if any
        if (postfields != NULL) {
            curl_easy_setopt(curl, CURLOPT_POSTFIELDS, postfields);
        }

        // enable redirect following, some providers have certain caldav
        // endpoinds on sub-paths
        curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);
        curl_easy_setopt(curl, CURLOPT_MAXREDIRS, num_redirects);

        res = curl_easy_perform(curl);

        long response_code = 0;
        curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &response_code);

        // free the memory used for the curl slist
        curl_slist_free_all(header);

        curl_easy_cleanup(curl);

        if (res != CURLE_OK) {
            tracepoint(diary, error_string, "curl_easy_perform() failed in caldav_req()", curl_easy_strerror(res));
            tracepoint(diary, debug_long, "curl_easy_perform() response code", response_code);
            free(caldav_resp.memory);
            return NULL;
        }
    }

    return caldav_resp.memory;
}

/* Return current user principal from CalDAV XML response */
char* parse_caldav_current_user_principal(char* xml) {
    char* xml_key_pos = strstr(xml, "current-user-principal");
    // this XML does not contain a user principal at all
    if (xml_key_pos == NULL) {
        return NULL;
    }

    //<D:current-user-principal>
    //<D:href>/caldav/v2/diary.in0rdr%40gmail.com/user</D:href>
    char* tok = strtok(xml_key_pos, "<"); // D:current-user-principal>
    if (tok != NULL) {
        tok = strtok(NULL, "<"); // D:href>/caldav/v2/test%40gmail.com/user
        tok = strstr(tok, ">"); // >/caldav/v2/test%40gmail.com/user
        tok++; // cut >
        //char* tok_end = strrchr(tok, '/');
        //*tok_end = '\0'; // cut /user
    }

    return tok;
}

/* Return home set uri from CalDAV home set XML response */
char* parse_home_set(char* xml) {
    char* homeset_needle = "calendar-home-set";

    char* homeset_pos = strstr(xml, homeset_needle);
    if (homeset_pos == NULL) {
        tracepoint(diary, debug, "parse_home_set() homeset_pos was null");
        return NULL;
    }
    tracepoint(diary, debug_string, "parse_home_set() homeset_pos", homeset_pos);

    char* href = strrstr(homeset_pos, "href>/");
    if (href != NULL) {
        tracepoint(diary, debug_string, "parse_home_set() href", href);
        href = strtok(href, "<");
        tracepoint(diary, debug_string, "parse_home_set() href strtok", href);
        if (href != NULL) {
            href = strchr(href, '>');
            href++; // cut >
        }
        return href;
    }
    return NULL;
}

/* Return calendar uri from CalDAV home set XML response */
char* parse_caldav_calendar(char* xml, char* calendar) {
    char displayname_needle[strlen(calendar) + strlen("displayname></") + 1];
    sprintf(displayname_needle, "displayname>%s</", calendar);

    char* displayname_pos = strstr(xml, displayname_needle);
    // this XML multistatus response does not contain the users calendar
    if (displayname_pos == NULL) {
        tracepoint(diary, debug, "parse_caldav_calenar() displayname_pos was null");
        return NULL;
    }
    tracepoint(diary, debug_string, "parse_caldav_calenar() displayname_pos", displayname_pos);

    // shorten multistatus response and find last hyperlink
    *displayname_pos= '\0';

    tracepoint(diary, debug_string, "parse_caldav_calenar() shortened multistatus response", xml);
    char* href = strrstr(xml, "href>/");
    if (href != NULL) {
        tracepoint(diary, debug_string, "parse_caldav_calenar() last href", href);
        href = strtok(href, "<"); // :href>/caldav/v2/aaa%40group.calendar.google.com/events/
        tracepoint(diary, debug_string, "parse_caldav_calenar() last href strtok", href);
        if (href != NULL) {
            href = strchr(href, '>');
            href++; // cut >
        }
        return href;
    }
    return NULL;
}

/* Upload event to CalDAV server */
void put_event(struct tm* date, const char* dir, size_t dir_size, char* calendar_uri, bool basicauth) {
    // get entry path
    char path[100];
    char* ppath = path;
    char* descr;
    long descr_bytes;

    fpath(dir, dir_size, date, &ppath, sizeof path);
    if (ppath == NULL) {
        tracepoint(diary, error_date, "Cannot get file path for entry", date);
        return;
    }

    FILE* fp = fopen(path, "r");
    if (fp == NULL) perror("Error opening file");

    fseek(fp, 0, SEEK_END);
    descr_bytes = ftell(fp);
    rewind(fp);

    size_t descr_label_size = strlen("DESCRIPTION:");
    size_t descr_size = descr_bytes + descr_label_size;
    descr = calloc(descr_size + 1, sizeof(char));
    if (descr == NULL) {
        perror("malloc failed");
        fclose(fp);
        return;
    }

    strcat(descr, "DESCRIPTION:");

    size_t items_read = fread(descr + descr_label_size, sizeof(char), descr_bytes, fp);
    if (items_read != descr_bytes) {
        perror("Error while reading in read_tokenfile()");
        fclose(fp);
        return;
    }

    descr[descr_size] = '\0';
    char* folded_descr = fold(descr);

    char uid[9];
    strftime(uid, sizeof uid, "%Y%m%d", date);

    char* ics = "BEGIN:VCALENDAR\n"
                "PRODID:-//diary\n" // todo: add version
                "VERSION:2.0\n"
                "CALSCALE:GREGORIAN\n"
                "BEGIN:VEVENT\n"
                "UID:%s\n"
                "DTSTART;VALUE=DATE:%s\n"
                "SUMMARY:%s\n"
                "%s\n"
                "END:VEVENT\n"
                "END:VCALENDAR";
    char postfields[strlen(ics) + strlen(folded_descr) + 100];
    sprintf(postfields, ics,
            uid,
            uid,
            uid, // todo: display first few chars of DESCRIPTION as SUMMARY
            folded_descr);

    tracepoint(diary, debug_string, "PUT text/calendar", postfields);

    strcat(calendar_uri, uid);
    strcat(calendar_uri, ".ics");
    char* response = caldav_req(date, calendar_uri, "PUT", postfields, 0, basicauth, "text/calendar", 0);
    fclose(fp);
    free(folded_descr);
    free(descr);

    if (response == NULL) {
        tracepoint(diary, error, "PUT request for event failed in put_event()");
    }

    // free memory allocated to store curl response
    free(response);
}

/*
* Sync with CalDAV server.
* Returns the answer char of the confirmation dialogue
* Returns 0 if neither local nor remote file exists.
* Otherwise, returns -1 on error.
*/
int caldav_sync(struct tm* date,
                 WINDOW* header,
                 WINDOW* cal,
                 int pad_pos,
                 const char* dir,
                 size_t dir_size,
                 bool confirm) {
    pthread_t progress_tid;

    char* info_txt;

    bool google_oauth_enabled = !(strcmp(CONFIG.google_clientid, "") == 0 || strcmp(CONFIG.google_clientid, "") == 0);
    bool basicauth_enabled = !(strcmp(CONFIG.caldav_username, "") == 0 || strcmp(CONFIG.caldav_password, "") == 0);

    if (strcmp(CONFIG.caldav_server, "") == 0 || strcmp(CONFIG.caldav_calendar, "") == 0 || !(google_oauth_enabled || basicauth_enabled)) {
        tracepoint(diary, debug_string, "CONFIG.caldav_server", CONFIG.caldav_server);
        tracepoint(diary, debug_string, "CONFIG.caldav_calendar", CONFIG.caldav_calendar);
        tracepoint(diary, debug_string, "CONFIG.caldav_username", CONFIG.caldav_username);
        tracepoint(diary, debug_string, "CONFIG.caldav_password", CONFIG.caldav_password);
        wclear(header);
        wresize(header, LINES, getmaxx(header));
        info_txt = "Missing sync parameters. Configure CalDAV server and credentials.\n"
                   "Press any key to continue.";
        mvwaddstr(header, 0, 0, info_txt);
        wrefresh(header);

        // accept any input to proceed
        noecho();
        wgetch(header);
        echo();
        wclear(header);
    }

    if (google_oauth_enabled) {
        // fetch existing API tokens from file
        char* tokfile = read_tokenfile();
        free(tokfile);

        if (strcmp(access_token, "") == 0) {
            // no access token exists yet, create new verifier
            char challenge[GOOGLE_OAUTH_CODE_VERIFIER_LENGTH + 1];
            random_code_challenge(GOOGLE_OAUTH_CODE_VERIFIER_LENGTH, challenge);

            // fetch new code with verifier
            char* code = get_oauth_code(challenge, header);
            if (code == NULL) {
                tracepoint(diary, error, "Error retrieving oauth code in caldav_sync()");
                return -1;
            }

            // get acess token using code and verifier
            tracepoint(diary, debug, "Fetching access token with code challenge");
            get_access_token(code, challenge, false);

            free(code);
        }
    }

    pthread_create(&progress_tid, NULL, show_progress, (void*)header);
    pthread_detach(progress_tid);

    char* principal_postfields = "<d:propfind xmlns:d='DAV:' xmlns:cs='http://calendarserver.org/ns/'>"
                                 "<d:prop><d:current-user-principal/></d:prop>"
                                 "</d:propfind>";


    // check if we can use the token from the tokenfile
    char* user_principal = caldav_req(date, CONFIG.caldav_server, "PROPFIND", principal_postfields, 0, basicauth_enabled, "application/xml", 1);
    tracepoint(diary, debug_string, "User principal", user_principal);

    if (user_principal == NULL) {
        // The principal could not be fetched
        tracepoint(diary, debug, "Unable to fetch principal");
        if (google_oauth_enabled) {
            // get new acess token with refresh token
            tracepoint(diary, debug, "Fetching access token with refresh token");
            get_access_token(NULL, NULL, true);
            // Retry request for event with new token
            user_principal = caldav_req(date, CONFIG.caldav_server, "PROPFIND", principal_postfields, 0, basicauth_enabled, "application/xml", 1);
        }
    }

    if (user_principal == NULL) {
        pthread_cancel(progress_tid);
        wclear(header);
        wresize(header, LINES, getmaxx(header));
        // TODO: this could also be wrong basicauth credentials
        info_txt = "Offline, corrupted or otherwise invalid OAuth2 credential tokenfile.\n"
                   "Go online or delete tokenfile '%s' and restart diary to retry.\n"
                   "Press any key to continue.";
        mvwprintw(header, 0, 0, info_txt, CONFIG.google_tokenfile);
        wrefresh(header);

        // accept any input to proceed
        noecho();
        wgetch(header);
        echo();

        wclear(header);
        return -1;
    }

    char* current_user_principal = parse_caldav_current_user_principal(user_principal);
    tracepoint(diary, debug_string, "Current user principal", current_user_principal);

    // extract host without path as basis for further api paths
    char *caldav_host; char* caldav_host_scheme;
    CURLU *h = curl_url();
    CURLUcode uc = curl_url_set(h, CURLUPART_URL, CONFIG.caldav_server, 0);
    uc = curl_url_get(h, CURLUPART_HOST, &caldav_host, 0);
    uc = curl_url_get(h, CURLUPART_SCHEME, &caldav_host_scheme, 0);
    tracepoint(diary, debug_string, "Caldav server host name", caldav_host);
    tracepoint(diary, debug_string, "Caldav server scheme/protocol", caldav_host_scheme);
    tracepoint(diary, debug_string, "Caldav server host name", caldav_host);
    if (!uc) {
        tracepoint(diary, error, "cur_url_get() failed in caldav_sync()");
    }

    // get the home-set of the user
    char uri[300];
    sprintf(uri, "%s://%s%s", caldav_host_scheme, caldav_host, current_user_principal);
    // free memory allocated by curl request
    free(user_principal);

    char* homeset_request = "<d:propfind xmlns:d='DAV:' xmlns:c='urn:ietf:params:xml:ns:caldav'>"
                            " <d:prop>"
                            "  <c:calendar-home-set />"
                            " </d:prop>"
                            "</d:propfind>";

    char* home_set = caldav_req(date, uri, "PROPFIND", homeset_request, 0, basicauth_enabled, "application/xml", 0);
    tracepoint(diary, debug_string, "Home set xml", home_set);
    if (home_set == NULL) {
        pthread_cancel(progress_tid);
        wclear(header);
        wresize(header, LINES, getmaxx(header));
        info_txt = "Error while fetching home set in caldav_sync()\n"
                   "Press any key to continue.";
        mvwaddstr(header, 0, 0, info_txt);
        wrefresh(header);

        // accept any input to proceed
        noecho();
        wgetch(header);
        echo();

        wclear(header);
        return -1;
    }

    // parse home set from xml response
    char* home_set_parsed = parse_home_set(home_set);
    sprintf(uri, "%s://%s%s", caldav_host_scheme, caldav_host, home_set_parsed);

    char* calendar_request = "<d:propfind xmlns:d='DAV:' xmlns:cs='http://calendarserver.org/ns/' xmlns:c='urn:ietf:params:xml:ns:caldav'>"
                             " <d:prop>"
                             "  <d:resourcetype />"
                             "  <d:displayname />"
                             "  <cs:getctag />"
                             "  <c:supported-calendar-component-set />"
                             " </d:prop>"
                             "</d:propfind>";

    // get calendar from home-set
    char* calendar_xml = caldav_req(date, uri, "PROPFIND", calendar_request, 1, basicauth_enabled, "application/xml", 0);
    tracepoint(diary, debug_string, "Calendar set xml", calendar_xml);

    // get calendar URI from the home-set
    char* calendar_href = parse_caldav_calendar(calendar_xml, CONFIG.caldav_calendar);

    char* xml_filter = "<c:calendar-query xmlns:d='DAV:' xmlns:c='urn:ietf:params:xml:ns:caldav'>"
                       "<d:prop><d:getetag/><c:calendar-data/></d:prop>"
                       "<c:filter><c:comp-filter name='VCALENDAR'>"
                       "<c:comp-filter name='VEVENT'>"
                       "<c:time-range start='%s' end='%s'/></c:comp-filter>"
                       "</c:comp-filter></c:filter></c:calendar-query>";

    char dstr_start[30];
    char dstr_end[30];
    char* format_start = "%Y%m%dT000000Z"; // start of day
    char* format_end = "%Y%m%dT235900Z"; // end of day
    strftime(dstr_start, sizeof dstr_start, format_start, date);
    strftime(dstr_end, sizeof dstr_end, format_end, date);

    char caldata_postfields[strlen(xml_filter) + strlen(dstr_start) + strlen(dstr_end)];
    sprintf(caldata_postfields, xml_filter,
            dstr_start,
            dstr_end);

    // fetch event for the cursor date
    sprintf(uri, "%s://%s%s", caldav_host_scheme, caldav_host, calendar_href);

    char* event = caldav_req(date, uri, "REPORT", caldata_postfields, 1, basicauth_enabled, "application/xml", 0);
    // todo: warn if multiple events,
    // multistatus has more than just one caldav:calendar-data elements
    // currently, the code below will just extract the first occurance of
    // DTSTAMP, UID and DESCRIPTION (from the first event)

    if (event == NULL) {
        // Event not found. The curl request probably failed due to one of the missing input parameters:
        // - Google Client ID
        // - Google Secret ID
        // - Google Calendar Name
        tracepoint(diary, debug, "Event not found");
        pthread_cancel(progress_tid);
        wclear(header);
        free(home_set);
        return -1;
    }

    // get path of entry
    char path[100];
    char* ppath = path;
    fpath(CONFIG.dir, strlen(CONFIG.dir), date, &ppath, sizeof path);

    time_t epoch_zero = 0;

    // assume local file does not exist
    struct tm* localfile_time = gmtime(&epoch_zero);

    // check last modification time of local time
    struct stat attr;
    if (stat(path, &attr) == 0) {
        // if local file does exists, remember utc time of file on disk
        localfile_time = gmtime(&attr.st_mtime);
    }

    time_t localfile_date = mktime(localfile_time);

    struct tm* remote_datetime;
    time_t remote_date = 0;
    long search_pos = 0;

    // check remote DTSTAMP of remote event
    char* remote_dtstamp = extract_ical_field(event, "DTSTAMP", &search_pos, false);
    char* remote_uid = extract_ical_field(event, "UID", &search_pos, false);
    tracepoint(diary, debug_string, "event", event);
    tracepoint(diary, debug_string, "DTSTAMP field", remote_dtstamp);

    // init remote date to 1970, assume remote file does not exist
    remote_datetime = gmtime(&epoch_zero);

    if (remote_dtstamp != NULL) {
        // if remote date does exist, set date to the correct timestamp
        strptime(remote_dtstamp, "%Y%m%dT%H%M%SZ", remote_datetime);
        free(remote_dtstamp);
    }

    remote_date = mktime(remote_datetime);

    tracepoint(diary, debug_string, "Remote last modified", ctime(&remote_date));
    tracepoint(diary, debug_string, "Local last modified", ctime(&localfile_date));
    double timediff = difftime(localfile_date, remote_date);
    tracepoint(diary, debug_double, "Time diff between local and remote mod time", timediff);

    char* rmt_desc;
    char dstr[16];
    int conf_ch = 0;

    if (timediff == 0) {
        tracepoint(diary, debug, "Local and remote files have equal timestamp or don't exist, giving up.");
        pthread_cancel(progress_tid);
        wclear(header);
        // free memory allocated to store curl response
        curl_free(caldav_host_scheme);
        curl_free(caldav_host);
        free(event);
        free(remote_uid);
        free(home_set);
        return 0;
    } else if (timediff > 0) {
        tracepoint(diary, debug, "Local file is newer. Uploading to remote.");
        if (remote_uid) {
            // purge any existing daily calendar entries on the remote side
            char event_uri[300];
            sprintf(event_uri, "%s://%s%s%s.ics", caldav_host_scheme, caldav_host, calendar_href, remote_uid);
            //sprintf(event_uri, "%s%s%s.ics", CONFIG.caldav_server, calendar_href, remote_uid);
            tracepoint(diary, debug_string, "Event URI for DELETE request", event_uri);
            char* response = caldav_req(date, event_uri, "DELETE", NULL, 0, basicauth_enabled, "", 0);
            free(response);
        }

        put_event(date, dir, dir_size, uri, basicauth_enabled);

        pthread_cancel(progress_tid);
        wclear(header);

    } else if (timediff < 0) {
        rmt_desc = extract_ical_field(event, "DESCRIPTION", &search_pos, true);

        if (rmt_desc == NULL) {
            tracepoint(diary, error, "Could not fetch description of remote event. Aborting sync.");
            pthread_cancel(progress_tid);
            wclear(header);
            curl_free(caldav_host_scheme);
            curl_free(caldav_host);
            free(event);
            free(remote_uid);
            free(home_set);
            return -1;
        }

        if (confirm) {
            // prepare header for confirmation dialogue
            curs_set(2);
            noecho();
            pthread_cancel(progress_tid);
            wclear(header);

            // ask for confirmation
            strftime(dstr, sizeof dstr, CONFIG.fmt, date);
            mvwprintw(header, 0, 0, "Remote event is more recent. Sync entry '%s' and overwrite local file? [(Y)es/(a)ll/(n)o/(c)ancel] ", dstr);
            conf_ch = wgetch(header);
        }

        if (conf_ch == 'y' || conf_ch == 'Y' || conf_ch == 'a' || conf_ch == '\n' || !confirm) {
            tracepoint(diary, debug, "Remote file is newer. Extracting description from remote.");
            char* i;

            // persist downloaded buffer to local file
            FILE* cursordate_file = fopen(path, "wb");
            if (cursordate_file == NULL) {
                perror("Failed to open cursor date file");
            } else {
                for (i = rmt_desc; *i != '\0'; i++) {
                    if (rmt_desc[i-rmt_desc] == 0x5C) { // backslash
                        switch (*(i+1)) {
                            case 'n':
                                fputc('\n', cursordate_file);
                                i++;
                                break;
                            case 0x5c: // preserve real backslash
                                fputc(0x5c, cursordate_file);
                                i++;
                                break;
                        }
                    } else {
                        fputc(*i, cursordate_file);
                    }
                }
            }
            fclose(cursordate_file);

            // add new entry highlight
            chtype atrs = winch(cal) & A_ATTRIBUTES;
            wchgat(cal, 2, atrs | A_BOLD, 0, NULL);
            prefresh(cal, pad_pos, 0, 1, ASIDE_WIDTH, LINES - 1, ASIDE_WIDTH + CAL_WIDTH);
        }

        echo();
        curs_set(0);
        free(rmt_desc);
    }

    // free memory allocated to store curl response
    curl_free(caldav_host_scheme);
    curl_free(caldav_host);
    free(event);
    free(remote_uid);
    free(home_set); // home_set was required for sprintfing calendar_href to event_uri
    return conf_ch;
}
